# Main function - Do Predictiveness curve estimation
predcurve <- function(dataset=NULL, d, markers, 
   riskFactors=NULL, nested=FALSE, link="logit", rho=NULL, 
   riskl=NULL, riskh=NULL, class=FALSE, ci=FALSE,  
   ylim=NULL, offset=.005, class_offset=.005, 
   plot_type = "predictiveness", densBW = 2, 
   nsamp=1000, boot_strata=NULL, cluster=NULL, level=95, boot_seed=51, 
   pdf=NULL, inputRisk = FALSE, missing="PerModel", addLowerTickLabs=FALSE) {

   #parmList <- c("dataset","d","markers","riskFactors","nested","link","riskl","riskh",
   #               "ci","rho","ylim","class","offset","class_offset",
   #               "plot_type","nsamp","cluster","level","boot_seed","densBW","pdf")

   ### DEFINE HELPER FUNCTIONS

   # Bootstrap CIs
   rocbsPred <- function(coefRowSet) {
     #Create CI matrix, with first col=SEs, second and third cols=CI
     nParam    <- 2 + length(riskFactors)
     pullParms <- "coefs"
     riskNames <- NULL
     if(!is.null(riskl)) {
       nParam <- nParam + 1
       riskNames <- c(riskNames,"riskLperc")
     }
     if(!is.null(riskh)) {
       nParam <- nParam + 1
       riskNames <- c(riskNames,"riskHperc")
     }    
     if(!is.null(riskl) & class) {
       if(plot_type != "cdf") {
         nParam <- nParam + 2
         riskNames <- c(riskNames,c("tprL","fprL"))
       } else {
         nParam <- nParam + 2
         riskNames <- c(riskNames,c("riskLpercCase","riskLpercCont"))
       }
     } 
     if(!is.null(riskh) & class) {
       if(plot_type != "cdf") {
         nParam <- nParam + 2
         riskNames <- c(riskNames,c("tprH","fprH"))
       } else {
         nParam <- nParam + 2
         riskNames <- c(riskNames,c("riskHpercCase","riskHpercCont"))
       }
     }

     # paramNames is no longer determined here since it will be different
     #    for each model when nested==TRUE.
     #   Also, the size of CImat doesn't need to be figured out before hand,
     #      (and the calc below would not be correct when nested==TRUE)
   #   paramNames <- c("Intercept","marker",riskFactors,riskNames)
   #   CImat <- matrix(nrow=nParam * nmark, ncol=3)

     pullParms <- c(pullParms,riskNames)

     set.seed(boot_seed)

     thetahat.star <- replicate(nsamp, do.onePred(pullParms))
     alpha <- 1 - (level/100)
     na.rm = FALSE
     CImat <-cbind(apply(thetahat.star, 1, function(z){sd(z,na.rm=na.rm)}),
                  t(apply(thetahat.star, 1, 
                  function(z){
                    quantile(z, c(alpha/2, (1 - alpha/2)),na.rm=na.rm)})))

     diffmatCI      = NULL
     diffmatClassCI = NULL
     npPvalue       = NULL
     npPvalueClass  = NULL
     if(nmark == 2 & !(is.null(riskl) & is.null(riskh))) {
       numrows = dim(thetahat.star)[1]
       # if we get here then coefRowSet is a STRICT subset of the total rows
       #    AND the remaining rows (the riskPercentiles, FPR, and TPRS) ...
       #    should split evenly into 2 matching groups
       keepRows = setdiff(1:numrows,rowSet)
       numKeep  = length(keepRows)/2
       keepSet1 = keepRows[1:numKeep]    
       keepSet2 = keepRows[-(1:numKeep)]
       diffMat = thetahat.star[keepSet1,] - thetahat.star[keepSet2,]
       if(numKeep > 1) {
         diffmatCI = cbind(apply(diffMat, 1, function(z){sd(z,na.rm=na.rm)}),
                  t(apply(diffMat, 1, 
                  function(z){quantile(z, c(alpha/2, (1 - alpha/2)),na.rm=na.rm)})))
         npPvalue = 2*apply(diffMat, 1, function(z){min(mean(1*(z<0)),mean(1*(z>0)))})

       } else {
         # exactly one of riskl and riskh are specified and class==FALSE
         diffmatCI = matrix(c(sd(diffMat,na.rm=na.rm),
                          quantile(diffMat, c(alpha/2, (1 - alpha/2)),na.rm=na.rm)),
                          nrow=1)
         npPvalue = 2*min(mean(1*(diffMat<0)),mean(1*(diffMat>0)))
       }
       if(class) {
         # if we get here, diffmatCI must split evenly in 3's (risk perc, TPR,FPR)
         topPart = numKeep/3
         diffmatClassCI = diffmatCI[topPart + (1:(2*topPart)),]
         # the reasons for the matrix() is to handle the nrow==1 case
         diffmatCI = matrix(diffmatCI[1:topPart,],nrow=topPart)
         npPvalueClass = npPvalue[topPart + (1:(2*topPart))]
         npPvalue      = npPvalue[1:topPart]     
       }
     }

   #   V <- matrix(nrow=nParam,ncol=nParam,dimnames=list(paramNames, paramNames))
   #   for(i in 1:nParam) {
   #      for(j in i:nParam) {
   #         V[j,i] <- cov(thetahat.star[j,], thetahat.star[i,])
   #      }
   #   }

      return(list(CImat=CImat, pullParms=pullParms, 
                  diffmatCI=diffmatCI, diffmatClassCI=diffmatClassCI,
                  npPvalue=npPvalue, npPvalueClass=npPvalueClass)) #, V=V))
   }


   listNamesToVec <- function(aList,parmNames) {
      vOut <- NULL
      for(cName in parmNames) {
         vOut <- c(vOut,eval(parse(text=paste("aList$",cName,sep=""))))
      }
      vOut  
   }
   
   # do.onePred is the main function for boostrapping.  It calls
   #   utility functions for resampling and estimation and assembles the 
   #   estimates into the correct matrix.
   do.onePred <- function(pullParms) {
      
      if(!is.null(boot_strata)) {
         stratas <- getStrata(d_vector,0,bootStrataMat)
         resample.ctrl <- NULL
         resample.case <- NULL
         resample.cohort <- NULL
         for(s in 1:length(stratas)) {
            resampIds <- as.numeric(rownames(stratas[[s]]))
            # NOTE:  clusterUniqueMat, cListFull[[s]], strataCluster ARE ALL
            #   not used by getBstrapSample when cluster==FALSE
            strataCluster <- clusterIDs[resampIds]
            resample <- getBstrapSample(stratas[[s]],noccsamp, cluster,
                           clusterUniqueMat, cListFull[[s]], strataCluster)
            if(!noccsamp) {
               resample.ctrl <- c(resample.ctrl, resampIds[resample[[1]]])
               resample.case <- c(resample.case, resampIds[resample[[2]]])
            } else {
               resample.cohort <- c(resample.cohort, resampIds[resample])
            }
         }
      } else {
         resample <- getBstrapSample(dataset_ordered,noccsamp, cluster,
                                     clusterUniqueMat, cListFull)
         if(!noccsamp) {
            resample.ctrl <- resample[[1]]
            resample.case <- resample[[2]]
         } else {
            resample.cohort <- resample
         }
      }       
       
      #We have the resampled indices. Put together the new dataset
      if(!noccsamp) {
         newdata <- rbind(dataset_ordered[resample.ctrl,],
                          dataset_ordered[resample.case,])
      } else {
         newdata <-  dataset_ordered[resample.cohort,]
         orderD <- order(newdata[,1])
         newdata <- newdata[orderD,]
      }

      newdata <- data.frame(newdata)

      # Calculate/save each pred curve for each marker
      pCoutVec <- NULL
      for(markerI in 1:nmark) {
         pCout = getPredCurve(newdata, markerI, bootSample=TRUE)
         # note: the length of that returned from listNamesToVec will be different
         #     (for different markerI) when nested==TRUE
         pCoutVec = c(pCoutVec,listNamesToVec(pCout,pullParms))
      }
      
      pCoutVec
   }


   # Get predictiveness curve, and FPR, TPR, full percentiles at cutpoints
   #    markerI is the index (within markers) of the marker to fit ...
   #       (as named in the in_dataset)
   # FOR NOW, THE getModel PARAMETER IS NOT USED!
   #  bootSample == TRUE is used when rho is specified to say that the data came
   #    from a resampling, and therefore doesn't need to be re-weighted
   #    as it will already have the correct case-control composition.
   getPredCurve <- function(in_dataset, markerI, 
                            getModel = FALSE, bootSample = FALSE) {

      dName      = colnames(in_dataset)[1] 
      markerName = markers[markerI]
      
      if(!nested) {
         formulaStr <- paste(dName,"~",paste(c(markerName,riskFactors),collapse="+"),
                             sep="")
         varList <- c(dName,markerName,riskFactors) 
      } else {
         if(markerName == "baseline") {
            # we know in this case there must be covariates
            formulaStr <- paste(dName,"~",
                                paste(riskFactors,collapse="+"), sep="")
            varList <- c(dName,riskFactors) 
         } else {
            markerSub = markers[1:markerI]
            # note: the next statement will never make markerSub have 0 length
            #    because of the above IF statement
            if(markerSub[1] == "baseline") markerSub = markerSub[-1]
            formulaStr <- paste(dName,"~",
                               paste(c(markerSub,riskFactors),collapse="+"),
                               sep="")
            varList <- c(dName,markerSub,riskFactors) 
         }
      }

      # HANDLE MISSINGNESS, DROP ALL ROWS FROM CURRENT SET WITHOUT 
      #   ALL REQUIRED ROWS
      if(missing=="AllModels") {
         # NOTE: given that we use %in% below, it will not matter if some of 
         #    markers are already in varList
         varList <- c(varList,markers)  
      }      
      colList    <- which( colnames(in_dataset) %in% varList)
      whichVec   <- which(0==apply(in_dataset,1,function(x) sum(is.na(x))))
      in_dataset <- in_dataset[whichVec,]
      
      if(!inputRisk) {
        if(link=="probit") {
           regrModel <- glm(as.formula(formulaStr), data=in_dataset, 
                          family=binomial(probit))
         } else {
           regrModel <- glm(as.formula(formulaStr), data=in_dataset, 
                          family=binomial(logit))
         }
         fitVal    <- as.vector(regrModel$fitted.values)
      } else {
         fitVal =  eval(parse(text=
                        paste("in_dataset$",markers[markerI],sep="")))
      }
     
      reorder   <- order(fitVal)
      fitVal    <- fitVal[reorder]
      d         <- in_dataset[reorder,1]
      n         <- length(d) 
      perc      <- rank(fitVal,ties="max")/n      

      whichCase <- which(d==1)
      numCase   <- length(whichCase)
      whichCont <- which(d==0)
      numCont   <- length(whichCont)      

      # NOTE: noccsamp will be TRUE if inputRisk==TRUE
      if(!noccsamp) { 
         expit      <- function(x) exp(x)/(1+exp(x))
         fitVal     <- expit(as.vector(regrModel$linear.predictors[reorder]) + 
                             log(numCont/numCase * rho/(1-rho)))
         rankCaseM  <- rank(fitVal[whichCase],ties="max")
         rankContM  <- rank(fitVal[whichCont],ties="max")
         rankAll    <- rank(fitVal,ties="max")
         perc            <- rep(NA,n)
         perc[whichCase] <- rho * rankCaseM / numCase + 
                            (1-rho)*(rankAll[whichCase] - rankCaseM) / numCont 
         perc[whichCont] <- rho * (rankAll[whichCont] - rankContM) / numCase + 
                            (1-rho) * rankContM / numCont 
      } else {
         rankCaseM  <- rank(fitVal[whichCase],ties="max")
         rankContM  <- rank(fitVal[whichCont],ties="max")
      }

      rankAll   <- rank(fitVal,ties="min")
      rankCase  <- rank(fitVal[whichCase],ties="min")
      rankCont  <- rank(fitVal[whichCont],ties="min")

      tprAll            = rep(NA,n)
      fprAll            = rep(NA,n)
      tprAll[whichCase] = (1 + numCase - rankCase)/numCase
      tprAll[whichCont] = (numCase - rankAll[whichCont] + rankCont)/numCase
      fprAll[whichCont] = (1 + numCont - rankCont)/numCont
      fprAll[whichCase] = (numCont - rankAll[whichCase] + rankCase)/numCont

      riskHperc = NULL
      riskLperc = NULL
      tprH      = NULL
      tprL      = NULL
      fprH      = NULL
      fprL      = NULL
      riskHpercCase = NULL
      riskLpercCase = NULL
      riskHpercCont = NULL
      riskLpercCont = NULL      

      if(!is.null(riskh)) {
         whichV    <- which(fitVal > riskh)
         lWhichV <- length(whichV)
         if(lWhichV > 0) {
            riskHPos  <- min(whichV)
            #riskHperc <- perc[riskHPos]   # 1 - perc[riskHPos]
            riskHperc <- 1 - (lWhichV/n)	    
            tprH      <- tprAll[riskHPos]
            fprH      <- fprAll[riskHPos]
         } else {
            riskHperc <- 1
            tprH      <- 0
            fprH      <- 0
         }
         if(plot_type == "cdf") {
            if(lWhichV > 0) {
               whichV  <- (1:n)[-whichV]
               lWhichV <- length(whichV)
            } else {
               whichV <- 1:n
               lWhichV <- n
            }
            if(lWhichV > 0) {
               riskHpercCase = sum(d[whichV])
               riskHpercCont = (lWhichV - riskHpercCase)/numCont
               riskHpercCase = riskHpercCase/numCase
            } else {
               riskHpercCont = 0
               riskHpercCase = 0
            }
         }
      }

      if(!is.null(riskl)) {
         whichV <- which(fitVal <= riskl)
         if(length(whichV) > 0) {
            riskLPos  <- max(whichV)
            riskLperc <- perc[riskLPos]
            tprL      <- tprAll[riskLPos]
            fprL      <- fprAll[riskLPos]
         } else {
            riskLperc <- 0
            tprL      <- 1
            fprL      <- 1
         }
         if(plot_type == "cdf") {
            if(length(whichV) > 0) {
               riskLpercCase = sum(d[whichV])
               riskLpercCont = (length(whichV) - riskLpercCase)/numCont
               riskLpercCase = riskLpercCase/numCase
            } else {
               riskLpercCont = 0
               riskLpercCase = 0
            }	    
         }	
      }

      if(!inputRisk) {      
         coefs  <- regrModel$coefficients
      } else {
         coefs  <- 1  # NOTE: this is here to avoid a 0 length, it's just
                      #    a placeholder.
      } 

      listOut = list(row=reorder, fitted = fitVal, perc = perc, 
           fpr = fprAll, tpr = tprAll,
           riskHperc=riskHperc,riskLperc=riskLperc,
           tprH=tprH,tprL=tprL, fprH=fprH,fprL=fprL, coefs=coefs)
      if(plot_type!="predictiveness") {
         # NOTE: these are removed at the very end before returning from main function
         listOut$rankCaseM = rankCaseM
         listOut$rankContM = rankContM
         listOut$whichCase = whichCase
         listOut$whichCont = whichCont

         listOut$riskHpercCase = riskHpercCase
         listOut$riskLpercCase = riskLpercCase
         listOut$riskHpercCont = riskHpercCont
         listOut$riskLpercCont = riskLpercCont	 
      }
      listOut
   }


   checkErrorsPredCurve <- function() {
      if(length(unique(d_vector))!=2) cat("d must take on two values\n")
      else if(min(d_vector)!=0 | max(d_vector)!=1) cat("d must be 0/1\n")
      else if(!(link=="probit" | link=="logit"))
         cat("link option must be either probit or logit if specified\n")
      else if(!(missing=="AllModels" | missing=="PerModel"))
         cat("missing option must be either PerModel or AllModels if specified\n")
      else if(ifelse(is.null(ylim),FALSE,ylim < 0 | ylim > 1))
         cat("ylim option must be between 0 & 1, if specified\n")
      else if(ifelse(is.null(riskl),FALSE,riskl < 0 | riskl > 1))
         cat("riskl option must be between 0 & 1, if specified\n")
      else if(ifelse(is.null(riskh),FALSE,riskh < 0 | riskh > 1))
         cat("riskh option must be between 0 & 1, if specified\n")
      else if(ifelse(is.null(rho),FALSE,rho <= 0 | rho >= 1))
         cat("rho option must be between 0 & 1, exclusive, if specified\n")
      else if(ifelse(is.null(offset),FALSE,offset < 0 | offset > .02))
         cat("offset option must be between 0 & .02, if specified\n")
      else if(ifelse(is.null(class_offset),FALSE,class_offset < 0 | class_offset > 2))
         cat("class_offset option must be between 0 & 2, if specified\n")
      else if(plot_type != "predictiveness" & plot_type != "cdf" &
              plot_type != "density") 
         cat("plot_type must be one of (\"predictiveness\",\"cdf\",\"density\")\n")

      # Check sampling variability options
      else if(!is.numeric(nsamp) | (nsamp < 1 & nsamp!=0))
         cat("nsamp argument must be >1\n")

      # Not in Stata code. My own additions
      else if(!(class==TRUE | class==FALSE))
         cat("class option must be TRUE or FALSE, if specified\n")
      else if(!(ci==TRUE | ci==FALSE))
         cat("ci option must be TRUE or FALSE, if specified\n")
      else if(!(level>0 & level<100))
         cat("level option must be between 0 & 100, if specified\n")
      else if(0==prod(sapply(markers,function(x) !(x %in% riskFactors))))
         cat("markers and riskFactors should not contain any of the same variables\n")
 
      else return(FALSE)
      return(TRUE)
   }
  
   init01Plot <- function(xaxes=TRUE,plotYLim=c(0,1)) {
      x = c(0,1)
      y = x
      plot(y~x,type="n",axes=FALSE,frame.plot=TRUE,xlab="",ylab="",ylim=plotYLim)
      byY = .2
      if(plotYLim[2] > 1.4) byY = .5
      axis(2,at=seq(plotYLim[1],plotYLim[2],by=byY),padj=1)
      if(xaxes) {
         if(plot_type=="density") {
            axis(1,at=c(0,.2,.4,.6,.8,1),padj=0)
         } else {
            axis(1,at=c(0,.2,.4,.6,.8,1),labels=c(0,20,40,60,80,100),padj=0)
         }
      }
      abline(h=seq(plotYLim[1],plotYLim[2],by=byY),col="lightgrey")
      abline(v=seq(0,1,by=.2),col="lightgrey")
   }

   setupGraphics <- function() {
      height = ifelse(class,7,4)
      width  = ifelse(!(is.null(riskl) & is.null(riskh)) & plot_type!="density",6,4)
      if(is.null(pdf)) {
         if(.Platform$GUI == "AQUA") {
            quartz(width=width,height=height)
         } else {
            windows(width=width,height=height)
         }
      } else {
         pdf(file=pdf,width=width,height=height)
      }
      if(class) par(mfrow=c(2,1))
   }
   closeGraphics <- function() {
     if(!is.null(pdf)) {
       dev.off()
     } 
   }

   plotCIs <- function(x,y,col="black",lwd=.9) {

      barWidth = .025 

      if(length(y) ==2) {
         segments(x,y[1],x,y[2],col=col,lwd=lwd)
         segments(x-barWidth/2,y[1],x+barWidth/2,y[1],col=col,lwd=lwd)
         segments(x-barWidth/2,y[2],x+barWidth/2,y[2],col=col,lwd=lwd)
      } else {
         segments(x[1],y,x[2],y,col=col,lwd=lwd)
         segments(x[1],y-barWidth/2,x[1],y+barWidth/2,col=col,lwd=lwd)
         segments(x[2],y-barWidth/2,x[2],y+barWidth/2,col=col,lwd=lwd)
      }
   }
  
   sLines <- function(y,x,col, lty,lwd) {
      
      n <- length(y)
      newX <- rep(x,each=2)[-2*n]
      newY <- rep(y,each=2)[-1]
      
      lines(newY~newX,col=col,lty=lty,lwd=lwd)
   }
   

   # pType == 1, main plot
   # pType == 2, classification plot
   #  marker is the index
   plotFunc <- function(currModel,markerI,pType,ciMat=NULL) {

      #allcols = brewer.pal(max(nmark,3),"Set1") 
      # the colors below come from ... allcols = brewer.pal(9,"Set1")
      allcols = c("#E41A1C","#377EB8","#4DAF4A","#984EA3","#FF7F00", 
                  "#FFFF33","#A65628","#F781BF","#999999")
      coli    = allcols[markerI]      

      if(ci) {
         ciRow = 1
         nmarkMed = median(1:nmark)
         offsetMult = sign(markerI - nmarkMed) * ceiling(abs(markerI - nmarkMed))
         ciCol = coli  #"black"
      }

      if(plot_type == "predictiveness") {
         ylab="Risk"
         xlab="% Population Below Risk"
         #xlab="Risk Percentile"
      } else if(plot_type == "density"){
         ylab="frequency"
         xlab="Risk"	    
      } else if(plot_type == "cdf"){
         ylab="Probability"
         xlab="Risk"	  	    
      }

      if(1 %in% pType) {
         # ylim is ignored for plot_types "cdf" or "density" 
         if(is.null(ylim) | plot_type != "predictiveness") { 
            plotYLim = c(0,1)
            topLY    = .9
         } else {
            plotYLim = c(0,ylim)
            topLY    = .9*ylim
         }
         if(plot_type == "density"){
            ylim = max(density(currModel$fitted,bw=densBW*bw.nrd0(currModel$fitted),
                       from=0,to=1)$y)
            plotYLim = c(0,1.2*ylim)
         }

         par(mfg=c(1,1))

         if(markerI==1) {
           parmarvec =  c(3,3,3,1.8)	     
           if(class) parmarvec[1] = .3   # reduce space for x labels in top plot
           if(!(is.null(riskl) & is.null(riskh)) & plot_type != "density") 
              parmarvec[4] = 13 #12
           par(mar=parmarvec)
           init01Plot(!class,plotYLim=plotYLim)
           if(nmark == 1) {
             title(paste("marker: ", markers[1],sep=""),cex.main=.8) 
           } else {
             if(nested) {
               if(markers[1] == "baseline") {
                 title(paste("markers: ",
                   paste(c("Risk Factors Only",markers[-1]),collapse=", +"),sep=""),
                   cex.main=.8)
               } else {
                 title(paste("markers: ",
                   paste(markers,collapse=", +"),sep=""), cex.main=.8)         
               }
             } else {
              title(paste("markers: ",paste(markers,collapse=", "),sep=""),
                  cex.main=.8) 
             }
           }
           title(ylab=ylab,cex.lab=.9,line=2)
           if(!class) title(xlab=xlab,,cex.lab=.9,line=2)
         }
         if(plot_type == "predictiveness") {
           sLines(currModel$fitted,currModel$perc,col=coli, lty="solid",lwd=3)
         } else if(plot_type == "density"){
           lines(density(currModel$fitted,bw=densBW*bw.nrd0(currModel$fitted),
              from=0,to=1), col=coli, lty="solid",lwd=3)
         } else if(plot_type == "cdf"){
           lines(currModel$perc~currModel$fitted,col=coli, lty="solid",lwd=3)
         }			

         # do segments at key values
         if(!is.null(riskl)) {
            if(plot_type == "predictiveness") {
               segments(currModel$riskLperc, -1, currModel$riskLperc, riskl,
                        col="black", lty="dashed", lwd=1)
               segments( 2, riskl, currModel$riskLperc, riskl, col="black",
                        lty="dashed",lwd=1)
               axis(4,at=riskl,lwd.ticks=2,las=2,hadj=.4,cex.axis=.8)
            } else if(plot_type == "density") {
               abline(v=riskl,col="black",lty="dashed",lwd=1)
            } else if(plot_type == "cdf") {	     
               segments(2, currModel$riskLperc, riskl, currModel$riskLperc, 
                        col="black", lty="dashed", lwd=1)
               segments(riskl,-1,riskl,currModel$riskLperc,col="black",
                        lty="dashed",lwd=1)
            #   axis(4,at=round(currModel$riskLperc,2),lwd.ticks=2,
            #        las=2,hadj=.4,cex.axis=.8)
            }	       
         }
         if(!is.null(riskh)) {
            if(plot_type == "predictiveness") {
               segments(currModel$riskHperc, -1, currModel$riskHperc, riskh,
                        col="black", lty="dashed", lwd=1)
               segments(2, riskh, currModel$riskHperc, riskh, col="black",
                        lty="dashed",lwd=1)
               axis(4,at=riskh,lwd.ticks=2,las=2,hadj=.4,cex.axis=.8)
            } else if(plot_type == "density") {
               abline(v=riskh,col="black",lty="dashed",lwd=1)
            } else if(plot_type == "cdf") {	     
               segments(2, currModel$riskHperc, riskh, currModel$riskHperc, 
                        col="black", lty="dashed", lwd=1)
               segments(riskh,-1,riskh,currModel$riskHperc,col="black",
                        lty="dashed",lwd=1)
            #   axis(4,at=round(currModel$riskHperc,2),lwd.ticks=2,
            #        las=2,hadj=.4,cex.axis =.8)
            }
         }

         # do legend
         if(markerI== nmark) {
           legendText = markers
           if(nested) legendText = paste(c("",rep("+ ",nmark - 1)),markers)
           if(nmark > 1) {
              if(plot_type == "predictiveness") {
                 legend("topleft",inset=.05,col=allcols[1:nmark],legend=legendText,
                        lwd=3,bg="white",cex=.7)
              } else if(plot_type == "density"){
                 legend("topright",inset=.05,col=allcols[1:nmark],legend=legendText,
                        lwd=3,bg="white",cex=.7)  
              } else if(plot_type == "cdf"){
                 legend("bottomright",inset=.05,col=allcols[1:nmark],legend=legendText,
                        lwd=3,bg="white",cex=.7)
              }
           }
         }
       
         # text output on the right side
         if(!(is.null(riskl) & is.null(riskh)) & plot_type!="density") {
            # we output numbers on the right
            #    since the user has asked for riskl and/or riskh numbers
            s15 = "               "	    
            textOutput1 = paste(s15,s15,s15,s15,"\n",sep="")
            textOutput2 = ""
            textOutput3 = ""
            textOutput4 = ""
            if(!is.null(riskh)) {
               paste(textOutput1," \n",sep="")
               format = "                risk threshold (high) = %3.2f\n"
               format2 = "%26s high-risk %%: % 3.1f\n"
               if(markerI==1) {
                  textOutput1 = paste(textOutput1,sprintf(format,riskh),sep="")
                  textOutput1 = paste(textOutput1,
                     sprintf(format2,markers[markerI],100*(1-currModel$riskHperc)),
                     sep="")
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput1 = paste(textOutput1," \n",sep="")
                  }			      
               } else {
                  textOutput1 = paste(textOutput1," \n",sep="")
                  for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput1 = paste(textOutput1,
                                         sprintf(format2,markers[markerI],
                                         100*(1-currModel$riskHperc)),
                                         sep="")
                     } else {   
                        textOutput1 = paste(textOutput1," \n",sep="")
                     }
                  }
               }
            }
            if(!is.null(riskh) & class) {
               if(plot_type != "cdf") {
                  FPFformat = "%35s FPF: %5.3f\n"
                  TPFformat = "%35s TPF: %5.3f\n"
                  cTPR      = currModel$tprH
                  cFPR      = currModel$fprH
               } else {
                  FPFformat = "%32s Cont %%: %5.3f\n"
                  TPFformat = "%32s Case %%: %5.3f\n"
                  cTPR      = currModel$riskHpercCase
                  cFPR      = currModel$riskHpercCont
               }
               if(markerI==1) {
                  textOutput2 = sprintf(TPFformat,markers[markerI],cTPR)
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput2 = paste(textOutput2," \n",sep="")
                  }
                  textOutput2 = paste(textOutput2,
                                 sprintf(FPFformat,markers[markerI],cFPR),
                                 sep="")
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput2 = paste(textOutput2," \n",sep="")
                  }
               } else {
                  textOutput2 = ""
                  for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput2 = paste(textOutput2,
                                        sprintf(TPFformat,markers[markerI],cTPR),
                                        sep="")
                     } else {
                        textOutput2 = paste(textOutput2," \n",sep="")
                     }
                  }
                  for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput2 = paste(textOutput2,
                                        sprintf(FPFformat,markers[markerI],cFPR),
                                        sep="")
                     } else {
                        textOutput2 = paste(textOutput2," \n",sep="")
                     }
                  }
               }
            }
            if(!is.null(riskl)) {
               format = "                 risk threshold (low) = %3.2f\n"
               format2 = "%27s low-risk %%: %3.1f\n"
               textOutput3 = "\n\n"	      
               if(markerI==1) {
                  textOutput3 = paste(textOutput3,sprintf(format,riskl),sep="")
                  textOutput3 = paste(textOutput3,
                         sprintf(format2,markers[markerI],100*currModel$riskLperc),
                         sep="")
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput3 = paste(textOutput3," \n",sep="")
                  }
               } else {
                  textOutput3 = paste(textOutput3," \n",sep="")
                  if(nmark > 1) for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput3 = paste(textOutput3,
                          sprintf(format2,markers[markerI],100*currModel$riskLperc),
                          sep="")
                     } else {
                        textOutput3 = paste(textOutput3," \n",sep="")
                     }
                  }
               }
            }
            if(!is.null(riskl) & class) {
               if(plot_type != "cdf") {
                  FPFformat = "%35s FPF: %5.3f\n"
                  TPFformat = "%35s TPF: %5.3f\n"
                  cTPR      = currModel$tprL
                  cFPR      = currModel$fprL
               } else {
                  FPFformat = "%32s Cont %%: %5.3f\n"
                  TPFformat = "%32s Case %%: %5.3f\n"
                  cTPR      = currModel$riskLpercCase
                  cFPR      = currModel$riskLpercCont
               }
               if(markerI==1) {
                  textOutput4 = sprintf(TPFformat,markers[markerI],cTPR)
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput4 = paste(textOutput4," \n",sep="")
                  }
                  textOutput4 = paste(textOutput4,
                                      sprintf(FPFformat,markers[markerI],cFPR),
                                      sep="")
                  if(nmark > 1) for(ii in 2:nmark) {
                     textOutput4 = paste(textOutput4," \n",sep="")
                  }
               } else {
                  textOutput4 = ""
                  for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput4 = paste(textOutput4,
                                        sprintf(TPFformat,markers[markerI],cTPR),
                                        sep="")
                     } else {
                        textOutput4 = paste(textOutput4," \n",sep="")
                     }
                  }
                  for(ii in 1:nmark) {
                     if(ii==markerI) {
                        textOutput4 = paste(textOutput4,
                                        sprintf(FPFformat,markers[markerI],cFPR),
                                        sep="")
                     } else {   
                        textOutput4 = paste(textOutput4," \n",sep="")
                     }
                  }
               }
            }
            text(x=1.47,y=topLY,pos=1,cex=.8,
                 paste(textOutput1,textOutput2,textOutput3,textOutput4,sep=""),
                 xpd=TRUE) 
         }  #  if(!(is.null(riskl) & is.null(riskh)))
      }     #  if(1 %in% pType) 

      ###############
      ###############
      if(2 %in% pType) {
         ###########
         # do the plot setup	  
         par(mfg=c(2,1))
         if(markerI==1) {
            parmarvec =  c(3,3,.4,1.8)	     
            if(!(is.null(riskl) & is.null(riskh)) & plot_type!="density") 
               parmarvec[4] = 13  #12
            par(mar=parmarvec)
            plotYLim = c(0,1)
            if(plot_type == "density"){
               f1 <- currModel$fitted[currModel$whichCase] 
               ylim = max(density(f1,bw=densBW*bw.nrd0(f1), from=0,to=1)$y)
               f2 <- currModel$fitted[currModel$whichCont] 
               ylim2 = max(density(f2,bw=densBW*bw.nrd0(f2), from=0,to=1)$y)
               plotYLim = c(0,1.2*max(ylim,ylim2))
            }
            init01Plot(plotYLim=plotYLim)
            if(plot_type == "predictiveness") ylab="Positive Fraction"
            title(ylab=ylab,cex.lab=.9,line=2)
            title(xlab=xlab,cex.lab=.9,line=2)
         }

         ###########
         # do the actual plots

         ltyVec = c("solid","dashed")  # c("solid","solid")
         lwdVec = c(2,2)   #c(3,2)

         if(plot_type == "predictiveness") {	
            sLines(currModel$tpr,currModel$perc,col=coli, 
                  lty=ltyVec[1],lwd=lwdVec[1])
            sLines(currModel$fpr,currModel$perc,col=coli,
                  lty=ltyVec[2],lwd=lwdVec[2])
         } else if(plot_type == "density") {
            f1 <- currModel$fitted[currModel$whichCase] 
            lines(density(f1,bw=densBW*bw.nrd0(f1),from=0,to=1), 
                  col=coli, lty=ltyVec[1],lwd=lwdVec[1])
            f2 <- currModel$fitted[currModel$whichCont] 
            lines(density(f2,bw=densBW*bw.nrd0(f2),from=0,to=1), 
                  col=coli, lty=ltyVec[2],lwd=lwdVec[2])
         } else if(plot_type == "cdf") {
           y1 <- currModel$rankCaseM/(length(currModel$rankCaseM))     
           lines(y1~currModel$fitted[currModel$whichCase],
                 col=coli, lty=ltyVec[1],lwd=lwdVec[1])
           y2 <- currModel$rankContM/(length(currModel$rankContM))
           lines(y2~currModel$fitted[currModel$whichCont],
                 col=coli, lty=ltyVec[2],lwd=lwdVec[2])
         }

         ###########
         # do segments at key values (and side tick marks)
         if(!is.null(riskl)) {
           if(plot_type == "predictiveness") {
             yLev = min(currModel$tprL,currModel$fprL)
             segments(currModel$riskLperc, 2, currModel$riskLperc, yLev,
                        col="black", lty="dashed", lwd=1)
             segments(2, currModel$tprL, currModel$riskLperc, currModel$tprL, 
                        col="black", lty="dashed",lwd=1)
             segments(2, currModel$fprL, currModel$riskLperc, currModel$fprL,
                        col="black", lty="dashed",lwd=1)
             labels = FALSE
             if(addLowerTickLabs) 
               labels = round(c(currModel$tprL,currModel$fprL),2)
             axis(4,at=c(currModel$tprL,currModel$fprL),lwd.ticks=2,labels=labels)
           } else if(plot_type == "density") {
             abline(v=riskl, col="black", lty="dashed",lwd=1)
           } else if(plot_type == "cdf") {
             OnemtprL = currModel$riskLpercCase
             OnemfprL = currModel$riskLpercCont
             yLev = max(OnemtprL,OnemfprL)
             segments(riskl, -1, riskl, yLev, col="black", lty="dashed", lwd=1)
             segments(2, OnemtprL, riskl, OnemtprL, col="black", lty="dashed",lwd=1)
             segments(2, OnemfprL, riskl, OnemfprL, col="black", lty="dashed",lwd=1)
             axis(4,at=c(OnemtprL,OnemfprL),lwd.ticks=2,labels=FALSE)
           }
         }
         if(!is.null(riskh)) {
           if(plot_type == "predictiveness") {
             yLev = min(currModel$tprH,currModel$fprH) 
             segments(currModel$riskHperc, 2, currModel$riskHperc, yLev, 
                        col="black", lty="dashed", lwd=1)
             segments(2, currModel$tprH, currModel$riskHperc, currModel$tprH, 
                        col="black", lty="dashed",lwd=1)
             segments(2, currModel$fprH, currModel$riskHperc, currModel$fprH,
                        col="black", lty="dashed",lwd=1)
             labels = FALSE
             if(addLowerTickLabs) 
               labels = round(c(currModel$tprH,currModel$fprH),2)
             axis(4,at=c(currModel$tprH,currModel$fprH),lwd.ticks=2,labels=labels)
           } else if(plot_type == "density") {
             abline(v=riskh, col="black", lty="dashed",lwd=1)
           } else if(plot_type == "cdf") {
             OnemtprH = currModel$riskHpercCase
             OnemfprH = currModel$riskHpercCont
             yLev = max(OnemtprH,OnemfprH)
             segments(riskh, -1, riskh, yLev, col="black", lty="dashed", lwd=1)
             segments(2, OnemtprH, riskh, OnemtprH, col="black", lty="dashed",lwd=1)
             segments(2, OnemfprH, riskh, OnemfprH, col="black", lty="dashed",lwd=1)
             axis(4,at=c(OnemtprH,OnemfprH),lwd.ticks=2,labels=FALSE)
           }
         }

         ###########
         # do legend
         if(markerI == nmark) {
           legendText = markers
           if(nested) legendText = paste(c("",rep("+ ",nmark - 1)),markers)
            if(plot_type == "predictiveness") {
               ltext = c("TPF","FPF")
               if(nmark > 1) ltext = 
                 c(paste(legendText,"TPF"),paste(legendText,"FPF"))
               legend("bottomleft", inset=.05,col=rep(allcols[1:nmark],2), 
                      legend=ltext, lwd=rep(lwdVec,each=nmark),
                      lty=rep(ltyVec,each=nmark), bg="white",cex=.7)
            } else if(plot_type == "density") {
               ltext = c("Cases","Controls")
               if(nmark > 1) ltext = 
                 c(paste(legendText,"Cases"),paste(legendText,"Controls"))
               legend("topright",inset=.05,col=rep(allcols[1:nmark],2), 
                      legend=ltext, lwd=rep(lwdVec,each=nmark),
		      lty=rep(ltyVec,each=nmark), bg="white",cex=.7)
            } else if(plot_type == "cdf") {     
               ltext = c("Cases","Controls")
               if(nmark > 1) ltext = 
	         c(paste(legendText,"Cases"),paste(legendText,"Controls"))
               legend("bottomright",inset=.05,col=rep(allcols[1:nmark],2), 
                      legend=ltext, lwd=rep(lwdVec,each=nmark),
		      lty=rep(ltyVec,each=nmark), bg="white",cex=.7)
            }
         }	
      }
   }
   
   
   # pType == 1, main plot
   # pType == 2, classification plot
   #  markerI is the index
   plotFuncP <- function(currModel,markerI,pType,ciMat=NULL) {

     #allcols = brewer.pal(max(nmark,3),"Set1") 
     # the colors below come from ... allcols = brewer.pal(9,"Set1")
     allcols = c("#E41A1C","#377EB8","#4DAF4A","#984EA3","#FF7F00", 
                 "#FFFF33","#A65628","#F781BF","#999999")
     coli    = allcols[markerI]      

     if(ci) {
       ciRow = 1
       nmarkMed = median(1:nmark)
       offsetMult = sign(markerI - nmarkMed) * ceiling(abs(markerI - nmarkMed))
       ciCol = coli  #"black"
       if(is.vector(ciMat)) ciMat <- matrix(ciMat,nrow=1)
     }

     if(1 %in% pType) {
       if(!is.null(riskl)) {
         if(plot_type == "predictiveness") {
           if(ci) {
             plotCIs(ciMat[ciRow,3:4],riskl+offsetMult*offset,ciCol)
             ciRow = ciRow + 1		  
           }
           points(currModel$riskLperc, riskl, cex=.8,pch=19) 
         } else if(plot_type == "density") {
           # do nothing
         } else if(plot_type == "cdf") {
           if(ci) {
             plotCIs(riskl+offsetMult*offset,ciMat[ciRow,3:4],ciCol)
             ciRow = ciRow + 1		  
           }
           points(riskl,currModel$riskLperc, cex=.8,pch=19)
         }
       }
       if(!is.null(riskh)) {
         if(plot_type == "predictiveness") {
           if(ci) {
             plotCIs(ciMat[ciRow,3:4],riskh+offsetMult*offset,ciCol)
             ciRow = ciRow + 1		  
           }
           points(currModel$riskHperc, riskh,cex=.8,pch=19) 
         } else if(plot_type == "density") {
           # do nothing
         } else if(plot_type == "cdf") {
           if(ci) {
             plotCIs(riskh+offsetMult*offset,ciMat[ciRow,3:4],ciCol)
             ciRow = ciRow + 1		  
           }
           points(riskh,currModel$riskHperc, cex=.8,pch=19)
         }
       }
     }

     ###############
     ###############
     if(2 %in% pType) {
       if(!is.null(riskl)) {
         if(plot_type == "predictiveness") {
           if(ci) {
             plotCIs(currModel$riskLperc+offsetMult*class_offset,
                     ciMat[ciRow,3:4],ciCol)
             plotCIs(currModel$riskLperc+offsetMult*class_offset,
                     ciMat[ciRow,7:8],ciCol)
             ciRow = ciRow + 1		  
           }     
           points(currModel$riskLperc, currModel$tprL, cex=.8,pch=19)
           points(currModel$riskLperc, currModel$fprL, cex=.8,pch=19)
         } else if(plot_type == "density") {
           # do nothing
         } else if(plot_type == "cdf") {
           if(ci) {
             plotCIs(riskl+offsetMult*class_offset, ciMat[ciRow,3:4],ciCol)
             plotCIs(riskl+offsetMult*class_offset, ciMat[ciRow,7:8],ciCol)
             ciRow = ciRow + 1		  
           } 
           points(riskl, currModel$riskLpercCase, cex=.8,pch=19)
           points(riskl, currModel$riskLpercCont, cex=.8,pch=19)
         }
       }
       if(!is.null(riskh)) {
         if(plot_type == "predictiveness") {
           if(ci) {
             plotCIs(currModel$riskHperc+offsetMult*class_offset,
                     ciMat[ciRow,3:4],ciCol)
             plotCIs(currModel$riskHperc+offsetMult*class_offset,
                     ciMat[ciRow,7:8],ciCol)
             ciRow = ciRow + 1		  
           }
           points(currModel$riskHperc, currModel$tprH,cex=.8,pch=19)
           points(currModel$riskHperc, currModel$fprH,cex=.8,pch=19)
         } else if(plot_type == "density") {
           # do nothing
         } else if(plot_type == "cdf") {
           if(ci) {
             plotCIs(riskh+offsetMult*class_offset, ciMat[ciRow,3:4],ciCol)
             plotCIs(riskh+offsetMult*class_offset, ciMat[ciRow,7:8],ciCol)
             ciRow = ciRow + 1		  
           } 
           points(riskh, currModel$riskHpercCase, cex=.8,pch=19)
           points(riskh, currModel$riskHpercCont, cex=.8,pch=19)
         }  
       }
     }
   }   

   
   formatMat <- function(x,format) {
     if(is.matrix(x)) out = apply(x,2,function(x) sprintf(format,x))
     if(!is.matrix(x)) out = sprintf(format,x)
     out
   }
   slashMatrix<- function(m) {
     ncm = ncol(m)
     if(ncm <= 2) return(m)
     ncm = ncol(m)
     m2 <- matrix(as.matrix(rbind(m,m)),ncol=2*ncm)  
     m2[,2*(1:ncm)] <- "|"
     colnames(m2) <- rep(colnames(m),each=2)
     colnames(m2)[2*(1:ncm)]  <- ""
     rownames(m2)  <- rep("",dim(m2)[1])
     removeCol <- c(2,2*ncm)
     if(ci) {
       # the order of columns is 2 label columns, followed by a pt. est.
       #  followed by 2 CI cols... this removes
       if(ncm >5) removeCol <- c(removeCol,8,14)
       if(ncm <= 5) removeCol <- c(removeCol,8)
     }
     return(m2[,-removeCol])
   }   
   
   
   
   ################################################   
   ################################################   
   ################################################
   ### END OF HELPER FUNCTIONS DEFINITIONS

   
   ### START OF "MAIN"
   
   optionSetting = getOption("warn")
   options(warn=-1)
   
   # Assign general variables
   nmark <- length(markers)
   if(link=="") link <- "probit"
   if(nsamp==0) ci <- FALSE
 
   adjust <- ifelse(!is.null(riskFactors), TRUE, FALSE)
   
   noccsamp = TRUE
   if(inputRisk) {
     if(!is.null(rho)) {
       rho = NULL
       cat(sprintf("Forcing cohort sampling (ignoring rho) since inputRisk==TRUE.\n"))
     }
     if(nested) {
       nested=FALSE
       cat(sprintf("Forcing no nesting since inputRisk==TRUE.\n"))
     }
     if(!is.null(riskFactors)) {
       riskFactors=NULL
       cat(sprintf("Ignoring riskFactors since inputRisk==TRUE.\n"))
     }
   }
   if(!is.null(rho)) {
     noccsamp = FALSE
     if(link == "probit") {
       cat(sprintf("Forcing link = \"logit\" since rho is specified.\n"))
       link = "logit"
     }
   }

   #Do error checking and create working dataset
   #If dataset variable is specified
   if(!is.null(dataset)) {
      # Check that specified dataset is valid
      if(!exists(dataset))
         stop("Specified dataset does not exist\n")

      # Check that specified disease variable is valid
      if(missing(d)) stop("d is not specified\n")
      if(!(d %in% names(eval(parse(text=dataset)))) )
         stop("Specified disease variable does not exist\n")

      # Check that specified marker variables are valid
      if(sum(markers %in% names(eval(parse(text=dataset))))<length(markers) ) 
         stop("At least one of the specified test markers does not exist\n")

      # Disease and test markers exist - start creating working dataset
      working_dataset <- as.data.frame(eval(parse(
                                       text=paste(dataset,"$",d,sep=""))))

      for(i in 1:nmark)
         working_dataset <- cbind(working_dataset, 
                             eval(parse(text=paste(dataset,"$",markers[i],sep=""))))

      covNames = NULL

      # Check that specified riskFactors are valid. 
      #    If so, add to working dataset.
      if(!is.null(riskFactors)) {
         if(sum(riskFactors %in% names(eval(parse(text=dataset))))<length(riskFactors) ) 
            stop("At least one of the specified riskFactors variables does not exist\n")
         else {
            covNames = c(covNames,riskFactors)
            for(i in 1:length(riskFactors))
               working_dataset <- cbind(working_dataset, 
                                 eval(parse(text=paste(dataset,"$",riskFactors[i],sep=""))))
         }
      }
      # Check that specified cluster variables are valid. 
      #     If so, add to working dataset (after checking we haven't already 
      #     done so).
      if(!is.null(cluster)) {
         if(sum(cluster %in% names(eval(parse(text=dataset))))<length(cluster) ) 
            stop("At least one of the specified cluster variables does not exist\n")
         else {
            for(i in 1:length(cluster))
               if(!(cluster[i] %in% covNames)) {
                  covNames = c(covNames,cluster[i])
                  working_dataset <- cbind(working_dataset, 
                       eval(parse(text=paste(dataset,"$",cluster[i],sep=""))))
               }
         }
      }
      # Check that specified boot_strata variables are valid. 
      #     If so, add to working dataset (after checking we haven't already 
      #     done so).
      if(!is.null(boot_strata) & ci) {
         if(sum(boot_strata %in% names(eval(parse(text=dataset))))<length(boot_strata) ) 
            stop("At least one of the specified boot_strata variables does not exist\n")
         else {
            for(i in 1:length(boot_strata))
               if(!(boot_strata[i] %in% covNames)) {
                  covNames = c(covNames,boot_strata[i])
                  working_dataset <- cbind(working_dataset, 
                       eval(parse(text=paste(dataset,"$",boot_strata[i],sep=""))))
               }
         }
      }
      
      colnames(working_dataset) <- c(d,markers,covNames)
   } else {  #If dataset variable is not specified
      working_dataset_colnames <- NULL
      # Check that specified disease variable is valid
      splitVar <- strsplit(d,"$",fixed=TRUE)
      if(length(splitVar[[1]]) == 1) {
         if(!exists(d))
            stop("Specified disease variable does not exist\n")
         working_dataset_colnames <- d
         working_dataset <- as.data.frame(eval(parse(text=d)))
      }
      else {
         if(!((splitVar[[1]])[2] %in% names(eval(parse(text=(splitVar[[1]])[1])))) )
            stop("Specified disease variable does not exist\n")
         working_dataset_colnames <- (splitVar[[1]])[2]
         working_dataset <- as.data.frame(eval(parse(text=d)))
         d <- (splitVar[[1]])[2]
      } 

      # Check that specified marker variables are valid
      splitVar <- strsplit(markers,"$",fixed=TRUE)
      for(i in 1:length(markers)) {
         if(length(splitVar[[i]]) == 1) {
            if(!exists(markers[i]))
               stop("At least one of the specified test markers does not exist\n")
            working_dataset_colnames <- c(working_dataset_colnames, markers[i])
            working_dataset <- cbind(working_dataset, 
                               eval(parse(text=paste(markers[i],sep=""))))
         } else {
            if(!((splitVar[[i]])[2] %in% names(eval(parse(text=(splitVar[[i]])[1])))) )
               stop("At least one of the specified test markers does not exist\n")
            working_dataset_colnames <- c(working_dataset_colnames, 
                                          (splitVar[[i]])[2] )
            working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(markers[i],sep=""))))
            markers[i] <- (splitVar[[i]])[2]
         }
      }

      # Check that specified riskFactors markers specified are valid
      if(!is.null(riskFactors)) {
         splitVar <- strsplit(riskFactors,"$",fixed=TRUE)
         for(i in 1:length(riskFactors)) {
            if(length(splitVar[[i]]) == 1) {
               if(!exists(riskFactors[i]))
                  stop("At least one of the specified riskFactors variables does not exist\n")
               if(!(riskFactors[i] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c(working_dataset_colnames, riskFactors[i])
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(riskFactors[i],sep=""))))
               }
            }
            else {
               if(!((splitVar[[i]])[2] %in% names(eval(parse(text=(splitVar[[i]])[1])))) )
                  stop("At least one of the specified riskFactors variables does not exist\n")
               if(!((splitVar[[i]])[2] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c( working_dataset_colnames, 
                                              (splitVar[[i]])[2] )
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(riskFactors[i],sep=""))))
               }
               riskFactors[i] <- (splitVar[[i]])[2]
            } 
         }
      }
      # Check that specified cluster variables are valid
      if(!is.null(cluster)) {
         splitVar <- strsplit(cluster,"$",fixed=TRUE)
         for(i in 1:length(cluster)) {
            if(length(splitVar[[i]]) == 1) {
               if(!exists(cluster[i]))
                  stop("At least one of the specified cluster variables does not exist\n")
               if(!(cluster[i] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c(working_dataset_colnames, cluster[i])
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(cluster[i],sep=""))))
               }
            }
            else {
               if(!((splitVar[[i]])[2] %in% names(eval(parse(text=(splitVar[[i]])[1])))) )
                  stop("At least one of the specified cluster variables does not exist\n")
               if(!((splitVar[[i]])[2] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c(working_dataset_colnames, 
                                                (splitVar[[i]])[2])
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(cluster[i],sep=""))))
               }
               cluster[i] <- (splitVar[[i]])[2]
            }
         }
      }
      # Check that specified boot_strata variables are valid
      if(!is.null(boot_strata) & ci) {
         splitVar <- strsplit(boot_strata,"$",fixed=TRUE)
         for(i in 1:length(boot_strata)) {
            if(length(splitVar[[i]]) == 1) {
               if(!exists(boot_strata[i]))
                  stop("At least one of the specified boot_strata variables does not exist\n")
               if(!(boot_strata[i] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c(working_dataset_colnames, 
                                                boot_strata[i])
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(boot_strata[i],sep=""))))
               }
            }
            else {
               if(!((splitVar[[i]])[2] %in% names(eval(parse(text=(splitVar[[i]])[1])))) )
                  stop("At least one of the specified boot_strata variables does not exist\n")
               if(!((splitVar[[i]])[2] %in% working_dataset_colnames)) {
                  working_dataset_colnames <- c(working_dataset_colnames, 
                                             (splitVar[[i]])[2])
                  working_dataset <- cbind(working_dataset, 
                                     eval(parse(text=paste(boot_strata[i],sep=""))))
               }
               boot_strata[i] <- (splitVar[[i]])[2]
            }
         }
      }     
     
      colnames(working_dataset) <- working_dataset_colnames
   }

   #### First round of error checking done. Working dataset created.   
   dataset_ordered <- working_dataset

   # NOTE: handling of missing data was moved to getPredCurve
   #   IT NOW HAPPENS on each dataset (boostrap or original)
  # missingDataRows <- (which(is.na(dataset_ordered), arr.ind=TRUE))[,1]
  # if(length(missingDataRows) > 0) 
  #    dataset_ordered <- dataset_ordered[-missingDataRows,]

   d_vector <- eval(parse(text=paste("dataset_ordered$",d,sep="")))

   if(ci & is.null(riskl) & is.null(riskh)) {
      ci = FALSE
      cat("ci is ignored (neither riskl nor riskh were specified).\n")
   } 
  
   if(!ci & !is.null(boot_strata)) {
      boot_strata = NULL
      cat("boot_strata is ignored (since ci is FALSE).\n")
   }   
   
   if(nested) {
      if(is.null(riskFactors) & length(markers) <= 1) {
         nested = FALSE
         cat("nested is ignored (no riskFactors were specified and only 1 marker).\n")
      }
   }
   if(nested) {
      # we know nesting makes sense because of the previous check.
      if(!is.null(riskFactors)) {
         nmark = nmark + 1
         markers = c("baseline",markers)
      }
   }
  
   clusterMat <- NULL
   clusterUniqueMat <- NULL
   clusterIDs <- vector(length=(dim(dataset_ordered)[1]))
   if(!is.null(cluster)) {
      clusterMat <- matrix(nrow=(dim(dataset_ordered))[1], ncol=length(cluster))
      for(i in 1:length(cluster))
         clusterMat[,i] <- eval(parse(text=
                              paste("dataset_ordered","$",cluster[i],sep="")))
      clusterMat <- data.frame(clusterMat)
      colnames(clusterMat) <- cluster
      clusterUniqueMat <- unique(as.data.frame(cbind(d_vector,clusterMat)))
      colnames(clusterUniqueMat) <- c(d,cluster)
      for(s in 1:(dim(clusterUniqueMat)[1])) {
         currCluster <- as.data.frame(cbind(d_vector, clusterMat, 
                                      seq(1:(dim(dataset_ordered)[1]))))
         colnames(currCluster) <- c(d,cluster,"origDatasetID")

         for(k in 1:length(cluster)) {
            currCluster <- currCluster[which(as.numeric(eval(parse(
                            text=paste("currCluster$",cluster[k],sep="")))) == 
                            clusterUniqueMat[s,k+1]),]
         }
         currCluster <- currCluster[which(
                    eval(parse(text=paste("currCluster$",d,sep=""))) == 
                    eval(parse(text=paste("clusterUniqueMat$",d,sep="")))[s]),]
         clusterIDs[currCluster$origDatasetID] <- s
      }
   }


   bootStrataMat <- NULL
   if(!is.null(boot_strata)) {
      bootStrataMat <- matrix(nrow=(dim(dataset_ordered))[1],
                              ncol=length(boot_strata))
      for(i in 1:length(boot_strata))
         bootStrataMat[,i] <- eval(parse(
                          text=paste("dataset_ordered","$",boot_strata[i],sep="")))
      bootStrataMat <- data.frame(bootStrataMat)
      colnames(bootStrataMat) <- boot_strata
   }  
  
   # calculate cluster lists
   cListFull <- NULL
   if(!is.null(cluster)) {
      # NOTE: strata variables work differently than for comproc, roccurve, rocglm
      #   There, they were adjustment covariates, and we specified whether 
      #      we wanted to treat them as factors (and stratify on them)
      #   HERE, it's a separate variable list. 
      if(!is.null(boot_strata)) {
         cListFull <- list()
         # make sure this works!  (setting y to 0)
         stratas <- getStrata(d_vector,0,bootStrataMat)
         for(s in 1:length(stratas)) {
            resampIds <- as.numeric(rownames(stratas[[s]]))
            cListFull[[s]] <- makeClist(clusterIDs[resampIds])
         }
      } else {
         cListFull <- makeClist(clusterIDs)
      }
   }  



   # Proceed if no errors
   isError <- checkErrorsPredCurve()
   resultList <- NULL   
   
   if(!isError) {

      ################      
      # Write preliminary display output        
      cat("\n\n*******************************************************\n")
      cat(sprintf("%41s\n","Predictiveness curve calculation"))
      if(nested) cat(sprintf("%32s\n","for Nested Models"))
      cat("\n")
      if(nmark == 1) { 
         cat(sprintf("%27s %s\n","marker:",markers[1]))
      } else {
         if(!nested) {
            cat(sprintf("%27s %10s\n","markers:",markers[1]))
            for(markerI in 2:nmark) cat(sprintf("%27s %10s\n"," ",markers[markerI]))
	 } else {
            baselineStr = markers[1]
            if(markers[1] == "baseline") baselineStr = "Risk Factors Only"
            cat(sprintf("%25s %s\n","baseline model:",baselineStr))
            cat(sprintf("%25s %s %s\n","Model 2:","baseline +",markers[2]))
            if(nmark > 2) {
               for(markerI in 3:nmark) 
                  cat(sprintf("%22s %d: Model %d + %s\n","Model",markerI,markerI-1,
                      markers[markerI]))
            }
         }
      }
      cat("\n")

      ################ 
      ################ 
      ################      
      # Fit models (just point estimates).      
      currModel <- list()
      for(markerI in 1:nmark) {      
         currModel[[markerI]] <- getPredCurve(dataset_ordered, markerI)
      }
      ################ 
      ################ 
      ################ 

      ################      
      # Setup/write matrices for storage of estimates.  3 matrices:
      #    (1) coefficient matrix (may not be output EVER).
      #    (2)  overall risk percentile matrix (only output if riskh or riskl !=NULL)
      #    (3)  TPR and FPR matrix (only output when (2) is output AND class=TRUE)
      if(!nested) {
         if(inputRisk) {
           regCoefNames <- "NA"
         } else {
           regCoefNames <- c("(Intercept)","marker")
           if(!is.null(riskFactors)) regCoefNames <- c(regCoefNames,riskFactors)
         }
      } else {
         regCoefNames <- list()
         for(markerI in 1:nmark) {
            regCoefNames[[markerI]] <- names(currModel[[markerI]]$coefs)
         }
      }
      riskPercMatNames <- NULL
      if(!is.null(riskl)) riskPercMatNames <- c(riskPercMatNames,
                                                sprintf("%4.3f (low)",riskl))      
      if(!is.null(riskh)) riskPercMatNames <- c(riskPercMatNames,
                                                sprintf("%4.3f (high)",riskh))
      
      PCcoef <- NULL
      for(markerI in 1:nmark) {      
         PCcoef <- c(PCcoef,  currModel[[markerI]]$coefs)
      }
      PCcoef <- matrix(PCcoef)

      PCriskP <- NULL
      if(!(is.null(riskl) & is.null(riskh))) {      
        for(markerI in 1:nmark) {      
           if(!is.null(riskl)) PCriskP <- c(PCriskP, 
                                            100*currModel[[markerI]]$riskLperc)
           if(!is.null(riskh)) PCriskP <- c(PCriskP, 
                                            100*currModel[[markerI]]$riskHperc)
        }
        PCriskP <- matrix(PCriskP)
      }      

      PCclass <- NULL
      if(!(is.null(PCriskP)) & class) {
         if(plot_type != "cdf") {
            for(markerI in 1:nmark) {      
               if(!is.null(riskl)) PCclass <- rbind(PCclass, 
                      c(currModel[[markerI]]$tprL,c(currModel[[markerI]]$fprL)))
               if(!is.null(riskh)) PCclass <- rbind(PCclass, 
                      c(currModel[[markerI]]$tprH,c(currModel[[markerI]]$fprH)))
            }
         } else {
            for(markerI in 1:nmark) {      
               if(!is.null(riskl)) PCclass <- rbind(PCclass, 
                                       c(currModel[[markerI]]$riskLpercCase,
                                       c(currModel[[markerI]]$riskLpercCont)))
               if(!is.null(riskh)) PCclass <- rbind(PCclass, 
                                       c(currModel[[markerI]]$riskHpercCase,
                                       c(currModel[[markerI]]$riskHpercCont)))
            }
         }
      }

      if(!nested) {
         totParam      <- length(regCoefNames) + length(riskPercMatNames) + 
                                 ifelse(class,2*length(riskPercMatNames),0)
         pOffset       <- (0:(nmark-1))*totParam
         lregCoefNames <- rep(length(regCoefNames),nmark)
      } else {
         lregCoefNames <-  unlist(lapply(regCoefNames,length))
         totParam <-  lregCoefNames + length(riskPercMatNames) + 
                      ifelse(class,2*length(riskPercMatNames),0)
         pOffset  <- cumsum(c(0,totParam[-nmark]))     
      }

      ################      
      # Do boostrap CIs if requested.      
      if(ci) {
         rowSet  <- NULL
         for(markerI in 1:nmark) 
            rowSet <- c(rowSet,pOffset[markerI] + (1:lregCoefNames[markerI]))

         bstrapResults <- rocbsPred(rowSet)	    

         PCcoef  <- cbind(PCcoef,bstrapResults$CImat[rowSet,])
         ##
         if(!(is.null(riskl) & is.null(riskh))) {
            rowSet  <- NULL
            for(markerI in 1:nmark) 
               rowSet <- c(rowSet,pOffset[markerI] + lregCoefNames[markerI] + 
                            1:length(riskPercMatNames))
            if(length(rowSet) > 1) {
               PCriskP <- cbind(PCriskP, 100*bstrapResults$CImat[rowSet,])
            } else {
               PCriskP <- matrix(c(PCriskP, 100*bstrapResults$CImat[rowSet,]),nrow=1)          
            }
         }
         ##
         if(!(is.null(riskl) & is.null(riskh)) & class) {
            rowSet  <- NULL
            for(markerI in 1:nmark) 
               rowSet <- c(rowSet,pOffset[markerI] + lregCoefNames[markerI] +
                           length(riskPercMatNames) + 1:(2*length(riskPercMatNames)))
            ciMatPart <- bstrapResults$CImat[rowSet,]
            ciMatPart <- matrix(t(ciMatPart),ncol=6,byrow=T)
            if(nmark > 1 | length(riskPercMatNames) > 1) {
               PCclass <- cbind(PCclass[,1], ciMatPart[,1:3], 
                                PCclass[,2], ciMatPart[,4:6])
            } else {
               PCclass <- matrix(c(PCclass[,1], ciMatPart[,1:3], 
                                 PCclass[,2], ciMatPart[,4:6]), nrow=1)              
            }
         }
      }
      
      ################      
      # Create text version of matrices to write to display
      markerCoef  <- rep(markers, lregCoefNames)
      fullMarkerCoef <- markerCoef
      markerCoef[which(duplicated(markerCoef))] <- ""
      if(nested) {
        regCoefNames <- as.vector(unlist(regCoefNames))
      } 
      
      markerPM  <- rep(markers, each=length(riskPercMatNames))
      markerPM[which(duplicated(markerPM))] <- ""

      dropBSE = TRUE

      #########
      output1 <- cbind(markerCoef,regCoefNames,formatMat(PCcoef,"%7.3f"))
      if(ci) {
         colnames(output1) <- c("Marker", "Coef.", "Observed", 
                                 "Bootstrap Std. Err.", 
                                 paste("[", level,"% Conf.",sep=""), 
                                 paste("Interval", "]", sep=""))
      } else {
         colnames(output1) <- c("Marker", "Coef.", "Observed")
      }
      resOutput1 <- as.data.frame(PCcoef)
      names(resOutput1) <- colnames(output1)[-c(1,2)]
      if(nmark == 1) rownames(resOutput1) <- regCoefNames     
      if(nmark > 1) rownames(resOutput1) <- paste(fullMarkerCoef,regCoefNames)     
      if(dropBSE & ci) {
         whichBSE <- which(colnames(output1) == "Bootstrap Std. Err.")
         output1 <- output1[,-whichBSE]
      }
      rownames(output1) <- rep("",dim(output1)[1])     

      #########
      output2 <- NULL
      resOutput2 <- NULL
      testFullPop <- ""
      if(!(is.null(riskl) & is.null(riskh))) {
        if(length(riskPercMatNames) > 1 | length(markerPM) > 1) {
          output2 <- cbind(markerPM,riskPercMatNames,formatMat(PCriskP,"%5.1f"))
        } else {
          output2 <- matrix(c(markerPM,riskPercMatNames,
                              formatMat(PCriskP,"%5.1f")), nrow=1)
        }    
        if(ci) {
          colnames(output2) <- c("Marker", "Risk Threshold", 
                                 "% Pop Below Risk", #"Risk Percentile", 
                                 "Bootstrap Std. Err.", 
                                 paste("[", level,"% Conf.",sep=""), 
                                 paste("Interval", "]", sep=""))
        } else {
          colnames(output2) <- c("Marker", "Risk Threshold", 
	      "% Pop Below Risk")  # "Risk Percentile")
        }
        resOutput2 <- as.data.frame(PCriskP)
        names(resOutput2) <- colnames(output2)[-c(1,2)]
        if(nmark == 1) rownames(resOutput2) <- riskPercMatNames
        if(nmark > 1) rownames(resOutput2) <- as.vector(
                          sapply(markers,function(x) paste(x,riskPercMatNames)))

        if(ci & nmark == 2) 	{	      
          # add difference matrix
          numrow = length(riskPercMatNames)
          # note: numrow >= 1 if we get here.
          # note: we subtract the point estimate, the SE, and the 2 CIs
          #    but we overwrite the SE and the 2 CIs below.
          diffPtEsts = PCriskP[1:numrow,] - PCriskP[numrow + (1:numrow),]
          if(numrow > 1) {
            diffMat = cbind(c("difference",rep("",numrow-1)),riskPercMatNames,
                             formatMat(diffPtEsts,"%5.1f"))
          } else {
            diffMat = matrix(c("difference",riskPercMatNames, 
                              formatMat(diffPtEsts,"%5.1f")),nrow=1)
            diffPtEsts = matrix(diffPtEsts,nrow=1)
          }
          diffMat[,c(4:6)] = formatMat(100*bstrapResults$diffmatCI,"%5.1f")
          output2 <- rbind(output2,"",diffMat)

          testFullPop = "\n"
          for(testI in 1:numrow) {
            # .01 is because we multiplied point estimates by 100 above
            zstat =  .01 * diffPtEsts[testI,1] / bstrapResults$diffmatCI[testI,1]
            pval =  2*(1-pnorm(abs(zstat)))
            testFullPop = paste(testFullPop, 
              "test of Ho: 0 = Difference at Risk Threshold ",
              riskPercMatNames[testI],  
              sprintf("\n (Normal Bootstrap) z = % 4.2f,  p = %3.2f, ",zstat,pval),
              sprintf("(NP Bootstrap) p = %3.2f\n",bstrapResults$npPvalue[testI]),
              sep="")
          }
        } 

        if(dropBSE & ci) {
          #  argh!!! I hate the fact that R will collapse matrix into vectors
          #    when the slice you take is one dimensional
          whichBSE <- which(colnames(output2) == "Bootstrap Std. Err.")
          cNames   <- colnames(output2)[-whichBSE]
          dimOut   <- dim(output2)[1]
          output2  <- matrix(output2[,-whichBSE],nrow=dimOut)
          colnames(output2) <- cNames
        }
        rownames(output2) <- rep("",dim(output2)[1])
      }

      #########
      output3 <- NULL
      resOutput3 <- NULL 
      testClass  <- ""
      if(!(is.null(riskl) & is.null(riskh)) & class) {
        # note: format string should match diffMat format string below
        if(length(riskPercMatNames) > 1 | length(markerPM) > 1) {
          output3 <- cbind(markerPM,riskPercMatNames, formatMat(PCclass,"%6.3f"))
        } else {
          output3 <- matrix(c(markerPM,riskPercMatNames, 
                              formatMat(PCclass,"%6.3f")), nrow=1)
        }

        if(ci) {
          colnames(output3) <- c("Marker", "Risk Threshold", "TPR", 
                                 "Bootstrap Std. Err.", 
                                 paste("[", level,"% Conf.",sep=""), 
                                 paste("Interval", "]", sep=""),
                                 "FPR", "Bootstrap Std. Err.", 
                                 paste("[", level,"% Conf.",sep=""), 
                                 paste("Interval", "]", sep=""))
          if(plot_type=="cdf") {
              colnames(output3)[c(3,7)] <- c("RiskPercCase","RiskPercCont")
          }
        } else {
          colnames(output3) <- c("Marker", "Risk Threshold", "TPR", "FPR")
          if(plot_type=="cdf") {
              colnames(output3)[3:4] <- c("RiskPercCase","RiskPercCont")
          }
        }

        resOutput3 <- as.data.frame(PCclass)
        names(resOutput3) <- colnames(output3)[-c(1,2)]
        if(nmark == 1) rownames(resOutput3) <- riskPercMatNames
        if(nmark > 1) rownames(resOutput3) <- as.vector(
                         sapply(markers,function(x) paste(x,riskPercMatNames)))
        if(ci & nmark == 2) 	{	      
          # add difference matrix
          numrow = length(riskPercMatNames)
          diffPtEsts = PCclass[1:numrow,] - PCclass[numrow + (1:numrow),]
          if(numrow > 1) {
            diffMat = cbind(c("difference",rep("",numrow-1)),riskPercMatNames,
                              formatMat(diffPtEsts,"%6.3f"))
          } else {
            diffMat = matrix(c("difference",riskPercMatNames,
                              formatMat(diffPtEsts,"%6.3f")),nrow=1)
            diffPtEsts = matrix(diffPtEsts,nrow=1)
          }
          bSres <- formatMat(bstrapResults$diffmatClassCI,"%6.3f")
          diffMat[,c(4:6,8:10)] = matrix(t(bSres),ncol=6,byrow=T) 
          output3 <- rbind(output3,"",diffMat)

          testClass = "\n"
          for(testI in 1:numrow) {
            bsOffset = (testI - 1) * 2
            zstat1 =  diffPtEsts[testI,1] / 
                            bstrapResults$diffmatClassCI[bsOffset + 1,1]
            pval1  =  2*(1-pnorm(abs(zstat1)))
            zstat2 =  diffPtEsts[testI,5] / 
                            bstrapResults$diffmatClassCI[bsOffset + 2,1]
            pval2  =  2*(1-pnorm(abs(zstat2)))
            testClass = paste(testClass, 
               "test of Ho: 0 = Difference in TPR at Risk Threshold ",
               riskPercMatNames[testI], "\n",  
               sprintf(" (Normal Bootstrap) z = % 4.2f,  p = %3.2f, ",zstat1,pval1),
               sprintf("(NP Bootstrap) p = %3.2f\n",
                        bstrapResults$npPvalueClass[bsOffset + 1]),
               "test of Ho: 0 = Difference in FPR at Risk Threshold ",
               riskPercMatNames[testI], "\n", 
               sprintf(" (Normal Bootstrap) z = % 4.2f,  p = %3.2f, ",zstat2,pval2),
               sprintf("(NP Bootstrap) p = %3.2f\n",
                        bstrapResults$npPvalueClass[bsOffset + 2]),
               sep="")
          }
        }
        if(dropBSE & ci) {
          #  argh!!! I hate the fact that R will collapse matrix into vectors
          #    when the slice you take is one dimensional
          whichBSE <- which(colnames(output3) == "Bootstrap Std. Err.")
          cNames   <- colnames(output3)[-whichBSE]
          dimOut   <- dim(output3)[1]
          output3  <- matrix(output3[,-whichBSE],nrow=dimOut)
          colnames(output3) <- cNames
        }	
	
        rownames(output3) <- rep("",dim(output3)[1])
      }
      
      
      ################      
      # Write numeric output to display      
      SDstring = ifelse(is.null(rho),"cohort","case-control")
      cat(sprintf("%30s %s\n","study design:",SDstring))
      if(!is.null(rho)) {
         cat(sprintf("%30s %2.1f\n", "population disease prevalence (rho):",rho)) 
         cat("\n")
         cat(sprintf("%40s\n","bootstrap percentile CI's based on sampling"))
         cat(sprintf("%30s\n","separately from cases and controls"))
      }
      cat("\n")
      cat(sprintf("%30s\n","risk estimation"))
      cat(sprintf("%30s %s\n","GLM link:",link))
      if(length(riskFactors) > 0) {  
          cat(sprintf("%30s %s\n","covariates:",riskFactors[1]))
          if(length(riskFactors) > 1) for(covarI in 2:length(riskFactors)) 
             cat(sprintf("%30s %s\n"," ",riskFactors[covarI]))
      }
      if(!(is.null(riskl) & is.null(riskh))) { 
         cat("\n*******************************************************\n")     
         #print(output1,quote=F)
         #cat("\n*******************************************************\n")
         print(slashMatrix(output2),quote=F)
	 cat(testFullPop)
         cat("\n*******************************************************\n")
         if(class) {
            print(slashMatrix(output3),quote=F)
            cat(testClass)
            cat("\n*******************************************************\n")
         }
      }	
      if(ci) cat(sprintf("# boostrap samples: %d\n",nsamp))
      if(!(is.null(riskl) & is.null(riskh)) & ci & nmark != 2) 
         cat(sprintf("(no testing was done since number of models != 2)\n"))
      cat("\n")      
      
      ################      
      # Output PLOTS    
      setupGraphics()
      numRows = length(riskPercMatNames)
      for(markerI in 1:nmark) {
         subPCriskP = NULL
         if(!(is.null(riskl) & is.null(riskh)) & ci) { 
            subPCriskP = PCriskP[(markerI -1)*numRows + (1:numRows),]/100
         }
         plotFunc(currModel[[markerI]],markerI,1,subPCriskP)
      }
      for(markerI in 1:nmark) {
         if(!(is.null(riskl) & is.null(riskh)) & ci) { 
            subPCriskP = PCriskP[(markerI -1)*numRows + (1:numRows),]/100
         }
         plotFuncP(currModel[[markerI]],markerI,1,subPCriskP)
      }   
     
      if(class) {
         for(markerI in 1:nmark) {
            subPCclass = NULL
            if(!(is.null(riskl) & is.null(riskh)) & ci) {
               subPCclass = PCclass[(markerI -1)*numRows + (1:numRows),]
            }
            plotFunc(currModel[[markerI]],markerI,2,subPCclass)
         }
         for(markerI in 1:nmark) {
            if(!(is.null(riskl) & is.null(riskh)) & ci) {
               subPCclass = PCclass[(markerI -1)*numRows + (1:numRows),]
            }
            plotFuncP(currModel[[markerI]],markerI,2,subPCclass)
         }
      }	  
      closeGraphics()
     
      ################      
      # Create output list      
      resultList = list(coefs =resOutput1, riskPerc =resOutput2, 
                        class =resOutput3, call = match.call())
      for(markerI in 1:nmark) {
         if(plot_type!="predictiveness") {
            currModel[[markerI]]$rankCaseM = NULL
            currModel[[markerI]]$rankContM = NULL
            currModel[[markerI]]$whichCase = NULL
            currModel[[markerI]]$whichCont = NULL

            currModel[[markerI]]$riskHpercCase = NULL
            currModel[[markerI]]$riskLpercCase = NULL
            currModel[[markerI]]$riskHpercCont = NULL
            currModel[[markerI]]$riskLpercCont = NULL
         }	  
         eval(parse(text=paste("resultList$",markers[markerI],
                               "_ptEst = currModel[[markerI]]",sep="")))
      }	       
   }
  
   options(warn=optionSetting)  
  
   invisible(resultList)  
}






