\name{predcurve}
\Rdversion{1.1}
\alias{predcurve}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Predictiveness Curve Estimation}
\description{
  Estimate a Predictiveness curve for a continuous disease marker(s) 
  or diagnostic test variables.
  Bootstrap confidence intervals for estimates are optionally included.  
  Basline risk factors can be accommodated in risk modelling.
}
\usage{
predcurve(dataset=NULL, d, markers, riskFactors=NULL, nested=FALSE,
   link="logit", rho=NULL, riskl=NULL, riskh=NULL, class=FALSE, ci=FALSE,  
   ylim=NULL, offset=.005, class_offset=.005, 
   plot_type = "predictiveness", densBW = 2, 
   nsamp=1000, boot_strata=NULL, cluster=NULL, level=95, boot_seed=51, 
   pdf=NULL, inputRisk = FALSE, missing="PerModel", addLowerTickLabs=FALSE) 
}
%- maybe also 'usage' for other objects documented here.

\arguments{
  \item{dataset}{optional character string specifying the name
      of the dataset to be used for analysis.}
  \item{d}{character string specifying the name of the 0/1
      (disease) outcome vector.}
  \item{markers}{(vector of) character string(s) specifying the 
      names of the test markers/variables.}
  \item{riskFactors}{optional (vector of) character string(s) specifying 
     variables other than the marker to be included in the risk model. 
     default is NULL}
  \item{nested}{logical specifying whether to create nested models.  
     When nested==TRUE, a risk-factor-only ("baseline") model will be done in 
     addition to the model with the marker + risk factors.  If multiple markers 
     are specified, the markers will also be nested.  The nested models will 
     be (1) Risk Factors only (if specified), (2) Risk Factors + first marker, 
     (3) Risk Factors + first and second marker, etc.  default is FALSE}
  \item{link}{character string specifying the risk model link function as
     "logit" (default) or "probit". "logit" link is forced when rho is 
     specified (see below). }
  \item{rho}{optional numeric specifying the population disease prevalence 
     which should be used with case-control data.  A cohort design is
     assumed by default.  When rho is specified: (1) overall risk distribution 
     is estimated as a weighted average of the case risk distribution and the 
     control risk distribution (2) "logit" link is forced}
  \item{riskl}{lower risk threshold at which to estimate risk percentiles 
     (and TPR/FPR when class==TRUE). default is NULL}
  \item{riskh}{upper risk threshold at which to estimate risk percentiles 
     (and TPR/FPR when class==TRUE). default is NULL}
  \item{class}{logical specifying whether to plot TPR & FPR curves in a 
     second panel.  default is FALSE}
  \item{ci}{logical specifying whether to include bootstrap confidence 
     intervals for risk, TPR, and FPR estimates. default is FALSE}
  \item{ylim}{optional upper limit for the y-axis scale for the 
     predictiveness plot (ignored for other plot types). default is 1}
  \item{offset}{numeric specifying the vertical CI offset from risk 
     thresholds used to avoid overlaying CIs from multiple markers. 
     default is .005 (0.5 percent of the the full scale).}
  \item{class_offset}{numeric specifying the horizontal CI offset for FPR and 
     TPR CI's.  default is .005 (0.5 percent of the the full scale) }
  \item{plot_type}{one of ("predictiveness","density","cdf").  Specifies 
     the type of plots to be created, predictiveness curves, density plots, or 
     CDF plots.  default is "predictiveness"}
  \item{densBW}{multiplier for the bandwidth used for smoothing when 
                density plots are requested.  default is 2}
  \item{nsamp}{numeric specifying the number of bootstrap samples to be drawn 
     for estimating sampling variability of parameter estimates. default is 1000.}
  \item{boot_strata}{optional (vector of) character string(s) specifying 
     variables that identify bootstrap resampling stratifications.  
     Bootstrap sampling will sample from each strata separately preserving
     the number of samples in each strata.}
  \item{cluster}{optional (vector of) character string(s) specifying variables 
     that identify bootstrap resampling clusters.  Bootstrap sampling will
     sample the entire cluster at once, including all data from each 
     sampled cluster.}
  \item{level}{numeric specifying the confidence level for CIs as a 
     percentage. default is 95}
  \item{boot_seed}{random seed to set at the begnining of bootstrap 
     sampling.  default is 51}
  \item{pdf}{optional character string specifying the filename for a PDF
      of the graphical output.  default == NULL which specifies that the
      graphical output is displayed on the screen.}
  \item{inputRisk}{logical specifying whether to treat the markers as risk
     estimates.  When TRUE, a risk model will not be fit but instead the
     provided markers will be treated as risk estimates.  When TRUE, the 
     parameters riskFactors, nested, link and rho will all be ignored (assumed = 
     default) since they involve making a risk estimate.  default = FALSE}
  \item{missing}{ one of ("PerModel","AllModels").  Specifies 
     how missingness will be handled.  "PerModel" missingness removes all
     rows in the dataset with missing values used in the current model.
     "AllModels" missingness removes all rows in the dataset with missing 
     values used in any model.  Values are removed after each bootstrap 
     sample is taken, and therefore each bootstrap sample may have a 
     different number of rows.  default = "PerModel"}
  \item{addLowerTickLabs}{logical specifying whether to include numerical
     labels (at the intersection points for riskh and riskl) in the 
     "class" plot for plot_type=="predictiveness".  default is FALSE}
}


\details{
  \code{predcurve} plots predictiveness curves, i.e. plots of estimated 
    disease risk vs. the risk distribution (empirical cdf).  Risk estimates are 
    based on a generalized linear binary model of disease risk as a function of 
    the specified marker variable(s).  Additional risk factors (specified 
    with riskFactors) can be included with the marker variable in the risk model.

    Risk percentiles and reference lines for specified high and/or low risk
    thresholds, riskh and riskl, are optionally included on the plot.

    An additional plot of the true and false positive fractions, TPR and FPR, as
    functions of the risk distribution is optionally included as are estimates 
    and reference lines corresponding to specified risk thresholds.

    Risk calculations assume a cohort sampling design by default.  A correction
    for case-control data will be employed if the population prevalence of 
    disease is specified with rho, and bootstrap samples for optional CI 
    calculation will be drawn separately from cases and controls.

    Testing of differences between models will be done whenever there are
    exactly 2 models (two markers or nested models), either riskl or riskh 
    is specified,  and ci==TRUE.  In such cases, the output will include 
    difference estimates & confidence intervals and tests of the hypothesis.  
    These tests are done in 2 ways: ("bootstrap normal") the SE of the bootstrap
    sample is used with the point estimate to form a Wald-statistic 
    ("non-parametric boostrap") the % of bootsrap samples which produced a 
    difference which was < 0 (or > 0 for negative point estimates).  

    Alternate graphical outputs (instead of the predictiveness curves) are 
    density plots and CDF plots.  These are specified with the plot_type 
    option.

    inputRisk allows for the input to be risk estimates.  When 
    inputRisk==TRUE, a risk model will not fit.  Instead the variables
    specified by markers will be treated as risk estimates.  In such cases,
    riskFactors, nested, link & rho all are ignored.

    NOTE: occaisionally fitting risk models to specific datasets will 
    produce warnings stating that the risk estimates were close to 0 OR 
    close to 1.  These warnings are all suppressed in this implementation.
}



\value{List containing function call information, estimates and 
confidence intervals. Returned list items include the following:
\item{coefs}{a dataframe of coefficients for the risk model(s) with bootstrap 
CI.}
\item{riskPerc}{a dataframe of risk percentage estimates with bootstrap 
CI. (raw form of the matrix written to the display)  NULL if neither riskh or riskl is specified. }
\item{class}{a dataframe of classification estimates with bootstrap 
CI. (raw form of the matrix written to the display)  NULL if class=FALSE 
or if neither riskh or riskl is specified. }
\item{call}{a string showing how the function was called.}
\item{"markerName"_ptEst}{the point estimates for each marker. 
       (one list for each marker)  The list contains: row (a vector 
showing the ordering of the other output vectors, eg fitted[1] corresponds 
to the row[1]'th row in the dataset), fitted (sorted estimated risks from the 
model) perc (sorted risk percentiles of the risk estimate), fpr (sorted false 
positive rates using that risk estimate as a cutoff),  tpr (sorted true 
positive rates using that risk estimate as a cutoff), riskHperc, riskLperc, tprH, tprL, fprH, fprL 
(estimates at the specified riskh and riskl), coefs (coefficients of the 
risk model)}
}



\references{
Pepe M.S., Feng Z., Huang Y., Longton G., Prentice R., Thompson I.M., 
Zheng Y. 2008. Integrating the predictiveness of a marker with its 
performance as a classifier.  \emph{American Journal of Epidemiology} 
\bold{167(3)}, 362--368.

Pepe M.S., Gu, W., Morris, D.E. 2010. The Potential of Genes and other Markers 
to Inform about Risk. \emph{Cancer Epidemiology, Biomarkers and 
Prevention} \bold{19(3)}, 655--665.
}


\author{
Daryl Morris, Fred Hutchinson Cancer Research Center and University of Washington, Seattle, WA.
\email{darylm@u.washington.edu}

Gary Longton, Fred Hutchinson Cancer Research Center, Seattle, WA.
\email{glongton@fhcrc.org}

Margaret Pepe, Fred Hutchinson Cancer Research Center and University of Washington, Seattle, WA.
\email{mspepe@u.washington.edu}
}


%\note{
%%  ~~further notes~~
%}

%% ~Make other sections like Warning with \section{Warning }{....} ~


\seealso{\code{\link{comproc}}, \code{\link{roccurve}}.}


\examples{
janssens <- read.table("http://labs.fhcrc.org/pepe/data/janssens_c.csv",
                       sep=",",header=T)

predcurve(dataset="janssens",d="disease", markers="logscr")
 
predcurve(dataset="janssens",d="disease", markers=c("logscr","bmi"))

predcurve(dataset="janssens",d="disease", markers=c("logscr","bmi"),riskh=.4)

predcurve(dataset="janssens",d="disease", markers=c("logscr","bmi"),
          riskh=.4,riskl = .1, class=TRUE)

predcurve(dataset="janssens",d="disease", markers="logscr", 
   riskFactors=c("age","hypertension","bmi","bruit","vascular","gender"), 
   link="logit", riskl=.1, riskh=.4, ci=TRUE, class=TRUE)

####### 
# example of nested marker models
m1 <- glm(disease ~ bmi +age +hypertension +bruit +vascular +gender,
          data=janssens, family=binomial)
janssens$m1Fit <- m1$linear.predictors  
m2 <- glm(disease ~ bmi +age +hypertension +bruit +vascular +gender +logscr,
          data=janssens, family=binomial)
janssens$m2Fit <- m2$linear.predictors  

predcurve(dataset="janssens", d="disease",markers=c("m1Fit","m2Fit"),
          riskh=.4,riskl = .1, class=TRUE, ci=TRUE)

####### 
# examples of alternate graphical outputs (cdf and density)
predcurve(dataset="janssens", d="disease",markers=c("m1Fit","m2Fit"),
          riskh=.4,riskl = .1, class=TRUE, plot_type="cdf",ci=TRUE)

predcurve(dataset="janssens", d="disease",markers=c("m1Fit","m2Fit"),
          riskh=.4,riskl = .1, class=TRUE, plot_type="density")

predcurve(dataset="janssens", d="disease",markers=c("m1Fit","m2Fit"),
          riskh=.4,riskl = .1, class=TRUE, plot_type="density",densBW=1)

####### 
# example using nested parameter (same as above except (1) no need for
#     fitting models outside of predcurve and (2) CI may differ 
#     since riskFactors are being included in the bootstrap)

predcurve(dataset="janssens",d="disease", markers="logscr", 
   riskFactors=c("bmi","age","hypertension","bruit","vascular","gender"), 
   link="logit", riskl=.1, riskh=.4, class=TRUE, ci=TRUE, nested=TRUE)

####### 
# examples to make figures from reproduce Pepe et al (2010). 
#  NOTE: using inputRisk=TRUE 
modelA <- read.table("http://labs.fhcrc.org/pepe/data/modelA.csv",
                       sep=",",header=T)
predcurve(dataset="modelA",d="disease",markers="risk",
          riskh=.2, class=TRUE, inputRisk=TRUE)
predcurve(dataset="modelA",d="disease",markers="risk",
          riskh=.2,class=TRUE, inputRisk=TRUE,plot_type="cdf")
predcurve(dataset="modelA",d="disease",markers="risk",
          riskh=.2,class=TRUE,inputRisk=TRUE,plot_type="density")


modelB <- read.table("http://labs.fhcrc.org/pepe/data/modelB.csv",
                       sep=",",header=T)

predcurve(dataset="modelB",d="disease",markers="risk",
          riskh=.2, class=TRUE, inputRisk=TRUE)
predcurve(dataset="modelB",d="disease",markers="risk",
          riskh=.2,class=TRUE, inputRisk=TRUE,plot_type="cdf")
predcurve(dataset="modelB",d="disease",markers="risk",
          riskh=.2,class=TRUE,inputRisk=TRUE,plot_type="density")


}


% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
%\keyword{ ~kwd1 }
%\keyword{ ~kwd2 }% __ONLY ONE__ keyword per line
